from os.path import expanduser
from pathspec import PathSpec
from pathspec.patterns import GitWildMatchPattern
from cryptography.fernet import Fernet
from json import loads
from functools import cache


@cache
def get_frost_config():
    with open(expanduser("~/.dfrost.json")) as f:
        return loads(f.read())


def get_storage_bucket():
    return get_frost_config()["storage_bucket"]


def get_encryption_key():
    return Fernet(get_frost_config()["encryption_key"].encode("utf-8"))


def get_sync_dir():
    return get_frost_config()["sync_dir"]


@cache
def get_request_list():
    lines = get_frost_config().get("request_list", list())
    return PathSpec.from_lines(GitWildMatchPattern, lines)


@cache
def get_publish_list():
    lines = get_frost_config().get("publish_list", list())
    return PathSpec.from_lines(GitWildMatchPattern, lines)



from importlib import import_module
from argparse import ArgumentParser
from dfrost.lib.log import info
from dfrost.lib.command import get_command_registry

import_module("dfrost.pull")
import_module("dfrost.push")
import_module("dfrost.generate")


def run():
    parser = ArgumentParser()

    subparsers = parser.add_subparsers(dest="_command")
    subparsers.required = True
    for name, (setup_fn, _command_fn) in get_command_registry().items():
        setup_fn(subparsers.add_parser(name))

    args = parser.parse_args()
    info(f"Invoked: {args._command}")
    get_command_registry()[args._command][1](args)


if __name__ == "__main__":
    run()



from os.path import join, getmtime, relpath
from os import walk
from cryptography.fernet import Fernet

from dfrost.lib.s3 import (
    FileDoesntExistError,
    get_file_state,
    upload_file,
)
from dfrost.lib.log import info, debug
from dfrost.lib.config import get_publish_list, get_sync_dir
from dfrost.lib.command import register


def setup_parser(parser):
    parser.add_argument("--dry-run", action="store_true")


@register("push", setup_fn=setup_parser)
def run(args):
    push(dry_run=args.dry_run)


def push(*, dry_run):
    for dirname, dirs, files in walk(get_sync_dir()):
        for fname in files:
            local_path = join(dirname, fname)
            s3_key = relpath(local_path, get_sync_dir())

            if not get_publish_list().match_file(s3_key):
                debug(f"File: {s3_key} skipped as it isn't published.")
                continue

            local_mtime = int(getmtime(local_path))

            try:
                state, mtime = get_file_state(s3_key)
                if local_mtime <= mtime:
                    debug(f"Skipping push of file: {s3_key} as it already exists")
                    continue
                info(f"File: {s3_key} is stale remotely and must be pushed")
                if not dry_run:
                    upload_file(s3_key, file_path=local_path, mtime=local_mtime)

            except FileDoesntExistError:
                info(f"File: {s3_key} is missing remotely and must be pushed")
                if not dry_run:
                    upload_file(s3_key, file_path=local_path, mtime=local_mtime)



from enum import Enum
from os import environ


class LogLevel(Enum):
    debug = 0
    info = 1
    warn = 2


def get_log_level():
    raw_level = environ.get("LOG_LEVEL", "WARN").lower()
    return LogLevel[raw_level]


def log(level, msg):
    if level.value < get_log_level().value:
        return
    print(f"[{level.name}] {msg}", flush=True)


def debug(msg):
    log(LogLevel.debug, msg)


def info(msg):
    log(LogLevel.info, msg)


def warn(msg):
    log(LogLevel.warn, msg)



from tempfile import TemporaryDirectory
from os.path import join
from contextlib import contextmanager


@contextmanager
def get_temp_fname():
    with TemporaryDirectory() as tmpdir:
        yield join(tmpdir, "tmp")



from dfrost.lib.config import get_encryption_key
from dfrost.lib.log import debug


def get_chunk_size():
    return 64 * 1024 * 1024


def encrypt_file(src, dst):
    debug(f"Encrypting file: {src} -> {dst}")
    fernet = get_encryption_key()
    with open(src, "rb") as src_f:
        with open(dst, "wb") as dst_f:
            while True:
                byte_data = src_f.read(get_chunk_size())
                if byte_data == b"":
                    break
                dst_f.write(fernet.encrypt(byte_data))


def decrypt_file(src, dst):
    debug(f"Decrypting file: {src} -> {dst}")
    fernet = get_encryption_key()
    with open(src, "rb") as src_f:
        with open(dst, "wb") as dst_f:
            while True:
                byte_data = src_f.read(get_chunk_size())
                if byte_data == b"":
                    break
                dst_f.write(fernet.decrypt(byte_data))



import itertools
from boto3 import client
from botocore.exceptions import ClientError
from enum import Enum, auto
from os import utime
from functools import cache
from dfrost.lib.log import debug
from dfrost.lib.config import get_storage_bucket
from dfrost.lib.crypt import decrypt_file, encrypt_file
from dfrost.lib.temp import get_temp_fname


class FileState(Enum):
    archived = auto()
    restoring = auto()
    restored = auto()


class FileStateInvalidError(Exception):
    pass


class FileDoesntExistError(Exception):
    pass


@cache
def get_s3_client():
    return client("s3")


def get_restore_days():
    return 7


def get_file_state(s3_key):
    try:
        debug(f"Checking state of file: {s3_key}")
        headers = get_s3_client().head_object(Bucket=get_storage_bucket(), Key=s3_key)[
            "ResponseMetadata"
        ]["HTTPHeaders"]
    except ClientError as err:
        if err.response["Error"]["Code"] == "404":
            raise FileDoesntExistError()
        raise

    mtime = int(headers["x-amz-meta-mtime"])
    if "x-amz-restore" not in headers:
        return FileState.archived, mtime
    if headers["x-amz-restore"] == 'ongoing-request="true"':
        return FileState.restoring, mtime
    return FileState.restored, mtime


def list_objects():
    for ix in itertools.count():
        debug(f"Fetching remote batch: {ix} of files")
        resp = get_s3_client().list_objects_v2(
            Bucket=get_storage_bucket(),
            **{"ContinuationToken": resp["NextContinuationToken"]} if ix > 0 else {},
        )
        for item in resp.get("Contents", list()):
            yield item["Key"]
        if not resp["IsTruncated"]:
            break


def restore_file(s3_key):
    try:
        debug(f"Restoring file: {s3_key}")
        get_s3_client().restore_object(
            Bucket=get_storage_bucket(),
            Key=s3_key,
            RestoreRequest={"Days": get_restore_days()},
        )
    except ClientError as err:
        if err.response["Error"]["Code"] == "RestoreAlreadyInProgress":
            raise FileStateInvalidError()
        if err.response["Error"]["Code"] == "NoSuchKey":
            raise FileDoesntExistError()
        raise


def upload_file(s3_key, *, file_path, mtime):
    with get_temp_fname() as fname:
        encrypt_file(file_path, fname)
        debug(f"Uploading file: {s3_key} from src: {fname}")
        get_s3_client().upload_file(
            fname,
            get_storage_bucket(),
            s3_key,
            ExtraArgs={
                "StorageClass": "DEEP_ARCHIVE",
                "Metadata": {"MTime": str(mtime)},
            },
        )


def download_file(s3_key, *, file_path, mtime):
    with get_temp_fname() as fname:
        try:
            debug(f"Downloading file: {s3_key} to dest: {fname}")
            get_s3_client().download_file(get_storage_bucket(), s3_key, fname)
        except ClientError as err:
            if err.response["Error"]["Code"] == "404":
                raise FileDoesntExistError()
            if err.response["Error"]["Code"] == "InvalidObjectState":
                raise FileStateInvalidError()
            raise
        decrypt_file(fname, file_path)
        utime(file_path, (mtime, mtime))



from os.path import join, exists, getmtime
from dfrost.lib.s3 import (
    FileDoesntExistError,
    FileState,
    FileStateInvalidError,
    download_file,
    list_objects,
    get_file_state,
    restore_file,
    upload_file,
)
from dfrost.lib.log import info, debug, warn
from dfrost.lib.config import get_request_list, get_sync_dir
from dfrost.lib.command import register


def setup_parser(parser):
    parser.add_argument("--dry-run", action="store_true")


@register("pull", setup_fn=setup_parser)
def run(args):
    pull(dry_run=args.dry_run)


def pull(*, dry_run):
    for s3_key in list_objects():

        if not get_request_list().match_file(s3_key):
            debug(f"File: {s3_key} skipped as it isn't requested.")
            continue

        try:
            state, mtime = get_file_state(s3_key)
            local_path = join(get_sync_dir(), s3_key)
            if exists(local_path) and int(getmtime(local_path)) >= mtime:
                debug(f"Skipping pull of file: {s3_key} as it already exists")
                continue

            if exists(local_path):
                info(f"File: {s3_key} is stale locally and must be pulled")
            else:
                info(f"File: {s3_key} is missing locally and must be pulled")

            if state == FileState.archived:
                info(f"File: {s3_key} is in an archived state and must be restored")
                if not dry_run:
                    restore_file(s3_key)
                continue
            if state == FileState.restoring:
                info(f"File: {s3_key} is in the process of being restored")
                continue
            if state == FileState.restored:
                info(f"File: {s3_key} is ready to be pulled")
                if not dry_run:
                    download_file(s3_key, file_path=local_path, mtime=mtime)

        except FileDoesntExistError:
            warn(f"File: {s3_key} unexpectedly disappeared")
            continue
        except FileStateInvalidError:
            warn(f"File: {s3_key} changed state unexpectedly")
            continue



from dfrost.lib.command import register
from cryptography.fernet import Fernet


@register("generate")
def run(args):
    generate()


def generate():
    print(Fernet.generate_key().decode("utf-8"), flush=True)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dfrost', 'dfrost.lib']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.20.51,<2.0.0',
 'cryptography>=36.0.1,<37.0.0',
 'pathspec>=0.9.0,<0.10.0']

entry_points = \
{'console_scripts': ['dfrost = dfrost.cli:run']}

setup_kwargs = {
    'name': 'dfrost',
    'version': '0.1.1',
    'description': '',
    'long_description': None,
    'author': 'tlonny',
    'author_email': 't@lonny.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from functools import cache


@cache
def get_command_registry():
    return dict()


def register(name, setup_fn=lambda par: par):
    def wrapper(command_fn):
        get_command_registry()[name] = (setup_fn, command_fn)
        return command_fn

    return wrapper
